shader{

	code= "
	uniform mat4		decalProjectionMatrix;
	uniform sampler2D	gBufferNormalTexture;
	uniform sampler2D	decalColorTexture;
	uniform sampler2D	decalNormalTexture;

	varying vec3		projectionDirection;
	varying vec3		tangent;
	varying vec3		binormal;
	
	void main(){
		
		//Calculate the screen space position of the pixel.
		gl_TexCoord[0].st /= gl_TexCoord[0].w;
		gl_TexCoord[0].st = gl_TexCoord[0].st * 0.5 + 0.5;

		//Obtain the world space position and geometry normal the pixel.
		vec3 position = reconstructPositionFromDepth(gl_TexCoord[0].st);
		vec3 geometryNormal = texture2D(gBufferNormalTexture, gl_TexCoord[0].st);

		vec3 decalTexCoord = (decalProjectionMatrix * vec4(position, 1.0)).xyz;
		decalTexCoord = decalTexCoord * 0.5 + 0.5;
		
		//Calculate how much is the surface facing the projection direction.
		float surfaceProjectionAngle = abs(dot(geometryNormal, projectionDirection));

		//Discard the pixel it it falls outside the projection matrix. The pixel is also discarded if the surface is not facing the decal to avoid precision errors.
		if(decalTexCoord.x < 0.0 || decalTexCoord.x > 1 || decalTexCoord.y < 0.0 || decalTexCoord.y > 1 || decalTexCoord.z < 0 || decalTexCoord.z > 1 || surfaceProjectionAngle < 0.1)
			discard;

		//Sample the transparency and normals from textures.
		vec4 transparency = texture2D(decalColorTexture, decalTexCoord.st).a;
		vec4 normalAndSpecularExponent = texture2D(decalNormalTexture, decalTexCoord.st);
		float specularExponent = 1.0;

		vec3 direction = vec4(geometryNormal, 0) * decalProjectionMatrix;
		if(abs(direction.y) >= 0.9)
		discard;

		//Project the tangent to the surface's plane.
		float d1 = dot(geometryNormal, tangent);
		tangent -= geometryNormal * d1;
		tangent = normalize(tangent);

		//Project the binormal to the surface's plane.
		float d2 = dot(geometryNormal, binormal);
		binormal -= geometryNormal * d2;
		binormal = normalize(binormal);

		//Build the tangent space matrix and transform the normal from tangent space to world space.
		mat3 tangentSpaceMatrix= mat3(tangent, binormal, geometryNormal);
		normalAndSpecularExponent.xyz = normalAndSpecularExponent.xyz * 2.0 - 1.0;
		vec3 worldNormal = normalAndSpecularExponent.xyz * transpose(tangentSpaceMatrix);
		worldNormal = normalize(worldNormal);

		//Output color and glossiness.
		gl_FragData[0]= vec4(worldNormal, 16.0);

		//Transparency is outputted to the second draw buffer and used as input to the blending
		//equation using dual-source blending to allow us to keep the alpha of the first buffer as data.
		gl_FragData[1]= vec4(0, 0, 0, transparency);
	}"

	includes{
		file= resources/shaders/engine/positionFromDepth.sh
	}

	semantics{
		gBufferNormalTexture = GBUFFER_GEOMETRY_NORMAL_TEXTURE
		decalProjectionMatrix = DEFERRED_DECAL_PROJECTION_MATRIX
		decalColorTexture = DEFERRED_DECAL_COLOR_TEXTURE
		decalNormalTexture = DEFERRED_DECAL_NORMAL_TEXTURE
	}
}