shader
{
	code= "
	uniform sampler2D	colorTexture;
	uniform sampler2D	normalTexture;
	uniform sampler2D	specularTexture;
	uniform float		specularRoughness;
	uniform float		specularIntensity;
	uniform float		specularFresnel;
	uniform float		alphaTestTreshold;
	uniform float		parallaxOcclusionMappingScale;
	uniform int			minParallaxOcclusionMappingSamples;
	uniform int			maxParallaxOcclusionMappingSamples;
	uniform vec3		cameraPosition;
	uniform float		sceneWetness;
	
	varying mat3		tangentSpaceMatrix;
	varying vec3		geometryNormal;
	varying float		linearDepth;
	varying vec3		position;
	varying vec3		tangentSpaceViewVector;
	varying float		terrainLayerWeight;


	vec2 reliefMapping(vec2 texcoord, vec2 ddx, vec2 ddy)
	{
		const float DEPTH_BIAS = 0.0001;
		
		tangentSpaceViewVector = normalize(tangentSpaceViewVector);

		//Select the number of samples based on the inclination of the view vector.
		int numSamples = mix(maxParallaxOcclusionMappingSamples,  minParallaxOcclusionMappingSamples, tangentSpaceViewVector.z);
		
		float cosine = dot(tangentSpaceViewVector,  vec3(tangentSpaceViewVector.xy, 0));
		float sine = sqrt(1 - cosine * cosine);
		float hipotenuse = parallaxOcclusionMappingScale / sine;

		vec2 samplingVector = tangentSpaceViewVector.xy * hipotenuse;
		vec2 samplingVectorStep = samplingVector / numSamples;
		float heightStep = 1.0 / numSamples;
		
		//Initialize the sampling data.
		vec2 samplingPoint = texcoord + samplingVector;
		vec2 previousSamplingPoint = samplingPoint;
		float height = 1.0;
		float previousHeight = 1.0;
		float previousSampleHeight = 1.0;

		//Perform the linear search.
		for(int i = 0; i < numSamples; i++)
		{
			//Obtain a filtered sample from the heightmap using the gradient information.
			float sampleHeight= texture2DGrad(normalTexture, samplingPoint, ddx, ddy).a;

			//Have we intersected the heightmap?
			if(height <= sampleHeight - DEPTH_BIAS)
			{
				//Now that we have intersected the heightfield let's estimate a sample position that
				//is a closer to the real heightmap surface. This is done using the current and last
				//height samples as a line/plane approximation of the heightmap surface and intersecting
				//the view vector against it. Check out the article from www.gamedev.net called 'A Closer
				//Look At Parallax Occlusion Mapping' for further details.

				float parallaxAmount = (previousSampleHeight - previousHeight) / ( heightStep + sampleHeight - previousSampleHeight);
				samplingPoint += samplingVectorStep * parallaxAmount;

				break;
			}
	
			//Store the information about the previous sample.
			previousSamplingPoint = samplingPoint;
			previousHeight = height;
			previousSampleHeight = sampleHeight;
			
			//Proceed to the next sample.
			samplingPoint -= samplingVectorStep;
			height -= heightStep;
		}

		return samplingPoint;
	}



	void applyWetness(inout vec4 color, inout float specularRoughness, inout float glossiness, inout float specularIntensity, inout float fresnel)
	{
		const float maximumWetnessPower = 1.5;
		const float waterSpecularRoughness = 0.02;
		const float waterGlossiness = 1.0;
		const float waterSpecularIntensity = 1.0;
		const float waterFresnel = 0.1;

		//Increase the contrast of the color in a non-linear way to simulate the wet look. We do this by raising
		//the color to a power which provides a good approximation to how the albedo changes for wet surfaces, check
		//the paper 'Determining Wet Surfaces From Dry' (page 3) to see a graphic of this phenomenon.
		color.rgb = pow(color.rgb, 1.0 + (maximumWetnessPower - 1.0) * sceneWetness);

		//Converge the material specular properties to the water specular properties based on the amount of wetness.
		specularRoughness = mix(specularRoughness, waterSpecularRoughness, sceneWetness);
		glossiness = mix(glossiness, waterGlossiness, sceneWetness);
		specularIntensity = mix(specularIntensity, waterSpecularIntensity, sceneWetness);
		fresnel = mix(fresnel, waterFresnel, sceneWetness);
	}


	
	void main()
	{
		//Alpha Testing ==========================================================
		#ifdef USE_ALPHA_TESTING

		//NOTE: for some reason, sampling the color texture with explicit partial derivatives and
		//discarding the pixel would cause undefined results and artifacts in the rendered image
		//even though no dynamic branching has occurred before getting the derivatives. Therefore,
		//to avoid those artifacts alpha testing is performed before the parallax occlusion maping
		//and without explicit derivatives.

		float alpha = texture2D(colorTexture, gl_TexCoord[0].st).a;

		//Discard pixel if alpha is bellow treshold.
		if(alpha <= alphaTestTreshold)
		{
			discard;
		}

		#endif
		//========================================================================


		//Parallax Occlusion Mapping =============================================

		//The Parallax Occlusion Mapping distorts the texture coordinates in an incoherent way which messes up the
		//screen space derivatives when the texture coordinates vary too much among adjacent pixels. This may cause
		//image artifacts due to the selection of incorrent mip-maps, so to avoid that we obtain the derivates before
		//any distortion of the texture coordinates and use them explicitely on all texture reads.
		vec2 ddx = dFdx(gl_TexCoord[0].st);
		vec2 ddy = dFdy(gl_TexCoord[0].st);

		#ifdef USE_PARALLAX_OCCLUSION_MAPPING

		gl_TexCoord[0].st = reliefMapping(gl_TexCoord[0].st, ddx, ddy);

		#endif
		//========================================================================

		vec4 color = texture2D(colorTexture, gl_TexCoord[0].st);
		float glossiness = texture2D(specularTexture, gl_TexCoord[0].st).x;
		vec3 normal = texture2D(normalTexture, gl_TexCoord[0].st).xyz * 2.0 - 1.0;
		vec3 worldSpaceNormal = normalize(normal * transpose(tangentSpaceMatrix));


		//Fresnel.
		vec3 viewVector = normalize(cameraPosition - position);
		float VdotN = max(dot(viewVector, worldSpaceNormal), 0.0);
		float fresnel = specularFresnel + (1.0 - specularFresnel) * pow(1.0 - VdotN + 0.0001, 5.0); //A small bias is applied to the calculation to avoid
																									//exponentiation of zero which leads do a denormalized float.

		applyWetness(color, specularRoughness, glossiness, specularIntensity, fresnel);

		//Combine the glossiness, specular intensity and fresnel into a single component.
		float specularMultiplier = glossiness * specularIntensity * fresnel;


		//Color and specular multiplier.
		gl_FragData[0]= vec4(color.rgb, 0.0);
	
		//World-space normal and specular roughness.
		gl_FragData[1]= vec4(worldSpaceNormal, specularRoughness);
		
		//World-space geometry normal.
		gl_FragData[2]= vec4(normalize(geometryNormal), specularMultiplier);
		
		//Linear depth.
		gl_FragData[3]= vec4(linearDepth, 1.0, 1.0, 1.0);
		
		//Position.
		gl_FragData[4]= vec4(position, 0.0);

		//Motion vector.
		gl_FragData[5]= vec4(calculateMotionVector(), 0, 0);


		//Blend the outputs by the terrain layer weight ==========================
		#ifdef USE_TERRAIN_LAYER_BLENDING

		gl_FragData[0] *= terrainLayerWeight;
		gl_FragData[1] *= terrainLayerWeight;
		gl_FragData[2] *= terrainLayerWeight;
		gl_FragData[3] *= terrainLayerWeight;
		gl_FragData[4] *= terrainLayerWeight;
		gl_FragData[5] *= terrainLayerWeight;

		#endif
		//========================================================================
	}"

	includes
	{
		file = resources/shaders/engine/deferredRendering/gBuffer/motionBlur.fs
	}

	semantics
	{
		colorTexture = MATERIAL_COLOR_TEXTURE
		specularTexture = MATERIAL_SPECULAR_TEXTURE
		normalTexture = MATERIAL_NORMAL_TEXTURE
		specularRoughness = MATERIAL_SPECULAR_ROUGHNESS
		specularIntensity = MATERIAL_SPECULAR_INTENSITY
		specularFresnel = MATERIAL_SPECULAR_FRESNEL
		alphaTestTreshold = MATERIAL_ALPHA_TEST_TRESHOLD
		parallaxOcclusionMappingScale = MATERIAL_PARALLAX_OCCLUSION_MAPPING_HEIGHT
		minParallaxOcclusionMappingSamples = MATERIAL_MIN_NUM_PARALLAX_OCCLUSION_MAPPING_SAMPLES
		maxParallaxOcclusionMappingSamples = MATERIAL_MAX_NUM_PARALLAX_OCCLUSION_MAPPING_SAMPLES
		cameraPosition = CAMERA_POSITION
		sceneWetness = SCENE_RAIN_WETNESS
	}
}