shader
{
	code= "
	uniform samplerCube	colorTexture;
	uniform samplerCube	normalTexture;
	uniform samplerCube	depthTexture;

	uniform vec3		lightPosition;
	uniform float		propagationVolumeWorldSpaceTexelSize;
	uniform float		propagationVolumeLightsWeight;
	uniform vec3		lightPropagationVolumePosition;
	uniform float		lightPropagationVolumeWidth;
	uniform float		lightPropagationVolumeHeight;
	uniform float		lightPropagationVolumeLength;
	uniform float		propagationVolumeHeightResolution;
	
	varying vec3		lightColor;
	varying vec3		lightDirection;
	
	void main()
	{
		//Read the VPL color, normal and position from the Reflective Shadow Map.
		vec3 samplingDirection = gl_Vertex.xyz;
		vec3 color = textureCube(colorTexture, samplingDirection).xyz;
		vec3 normal = textureCube(normalTexture, samplingDirection).xyz;
		float distance = textureCube(depthTexture, samplingDirection).x;
		vec3 position = lightPosition + samplingDirection * distance;
		
		//Half cell shifting of the position in the direction of normal and light to avoid self illumination.
		//Attention: actually, the length of the shifting is twice cell size and not half to reduce self illumination.
		vec3 shiftDirection= normalize(normal);
		position.xyz += shiftDirection * propagationVolumeWorldSpaceTexelSize * 2;
		
		//Multiply the color by the squared distance to compensate for the fact
		//that the concentration of VPLs decreases quadratically with distance.
		color *= distance * distance * 0.0001;

		//Injection=======================================================================================


		//Lay down the vertex on the unwrapped texture. This is done by finding to which layer the
		//vertex belongs to based on its Y coordinate and calculating an offset that is applied to
		//the vertex position.
		
		vec3 volumeDimensions= vec3(lightPropagationVolumeWidth, lightPropagationVolumeHeight, lightPropagationVolumeLength);
		vec3 positionInsideVolume= (position.xyz- (lightPropagationVolumePosition- volumeDimensions*0.5))/volumeDimensions;

		//Make sure the vertex is inside the layer, if not then
		//it will be ignored by assigning a black color to it.
		if(positionInsideVolume.x>= 0.0 && positionInsideVolume.x<= 1.0 && positionInsideVolume.y>= 0.0 && positionInsideVolume.y<= 1.0 && positionInsideVolume.z>= 0.0 && positionInsideVolume.z<= 1.0){
			float layer= floor(positionInsideVolume.y*propagationVolumeHeightResolution);
			position.x+= lightPropagationVolumeWidth*layer;
		}
		else
			color= vec3(0.0);

		gl_Position= gl_ModelViewProjectionMatrix * vec4(position, 1);
		lightColor= color;
		lightDirection= normal;
	}
	"
	
	semantics
	{
		lightPosition =	LIGHT_POSITION
		propagationVolumeWorldSpaceTexelSize = LIGHT_PROPAGATION_VOLUME_WORLDSPACE_TEXEL_SIZE
		propagationVolumeLightsWeight = LIGHT_PROPAGATION_VOLUME_LIGHTS_WEIGHT
		lightPropagationVolumePosition = LIGHT_PROPAGATION_VOLUME_POSITION
		lightPropagationVolumeWidth = LIGHT_PROPAGATION_VOLUME_WIDTH
		lightPropagationVolumeHeight = LIGHT_PROPAGATION_VOLUME_HEIGHT
		lightPropagationVolumeLength = LIGHT_PROPAGATION_VOLUME_LENGTH
		propagationVolumeHeightResolution = LIGHT_PROPAGATION_VOLUME_HEIGHT_RESOLUTION
	}
}