shader
{
	code= "
	uniform float	elapsedTime;

	struct EyeFocalAdaptationState
	{
		float circleOfConfusionMultiplier;
		float currentFocalPointDepth;
		bool isContracting;
	};

	layout (local_size_x = 1)								in;
	layout (std140, binding = 0) buffer						EyeFocalAdaptationStateBuffer
	{
		EyeFocalAdaptationState eyeFocalAdaptationState;
	};

	void main()
	{
		const float	CONTRACTION_SPEED = 0.1;
		const float	EXPANSION_SPEED = 1.0;
		const float FOCAL_POINT_SEARCH_RADIUS = 0.1;
		const int NUM_FOCAL_POINT_SEARCH_SAMPLES = 10;
		const float MAX_FOCAL_POINT_DIFFERENCE = 0.3;

		//Read the current state from the buffer.
		float circleOfConfusionMultiplier = eyeFocalAdaptationState.circleOfConfusionMultiplier;
		float currentFocalPointDepth = eyeFocalAdaptationState.currentFocalPointDepth;
		bool isContracting = eyeFocalAdaptationState.isContracting;

		//eyeFocalAdaptationState.circleOfConfusionMultiplier = 0.0;
		eyeFocalAdaptationState.circleOfConfusionMultiplier = min(eyeFocalAdaptationState.circleOfConfusionMultiplier + 0.01, 1.0);
		eyeFocalAdaptationState.currentFocalPointDepth = 0.0;
		eyeFocalAdaptationState.isContracting = false;

		//Calculate the new focal point by sampling the pixel depth at around the center of the screen.
		const float halfSearchRadius = FOCAL_POINT_SEARCH_RADIUS / 2.0;
		const float searchStep = FOCAL_POINT_SEARCH_RADIUS / (NUM_FOCAL_POINT_SEARCH_SAMPLES - 1);
		float newFocalPointDepth = 0.0;

		for(float y = 0.5 - halfSearchRadius; y <= 0.5 + halfSearchRadius; y += searchStep)
		{
			for(float x = 0.5 - halfSearchRadius; x <= 0.5 + halfSearchRadius; x += searchStep)
			{
				newFocalPointDepth += getPixelDepth(vec2(x, y));
			}
		}
				
		newFocalPointDepth /= (NUM_FOCAL_POINT_SEARCH_SAMPLES * NUM_FOCAL_POINT_SEARCH_SAMPLES); 

		//Calculate the difference between the new focal point and the previous.
		float focalPointDepthDifference = abs(newFocalPointDepth - currentFocalPointDepth);

		//If the new focal point is too different from the previous because
		//the camera moved then start expanding the depth of field so the
		//image is crisp while moving the camera.
		if(isContracting && focalPointDepthDifference > MAX_FOCAL_POINT_DIFFERENCE)
		{
			isContracting = false;
		}

		//Contract or expand the depth of field according to the current state.
		if(isContracting)
		{
			//Increase the multiplier smoothly converging it 1.
			circleOfConfusionMultiplier += (1.0 - circleOfConfusionMultiplier) * (1.0 - exp(-elapsedTime * CONTRACTION_SPEED));
		}
		else
		{
			//Decrease the multiplier smoothly converging it 0.
			circleOfConfusionMultiplier -= circleOfConfusionMultiplier * (1.0 - exp(-elapsedTime * EXPANSION_SPEED));


			//If the expansion finished then start contracting again.
			if(circleOfConfusionMultiplier < 0.1)
			{
				circleOfConfusionMultiplier = 0.0;
				isContracting = true;
			}
			else
			{
				//Keep the previous focal point to keep the depth of field consistent during the expansion.
				newFocalPointDepth = currentFocalPointDepth;
			}
		}
		
		//Store the updated state.
		eyeFocalAdaptationState.circleOfConfusionMultiplier = circleOfConfusionMultiplier;
		eyeFocalAdaptationState.currentFocalPointDepth = newFocalPointDepth;
		eyeFocalAdaptationState.isContracting = isContracting;
	}
	"

	includes
	{
		file = resources/shaders/engine/voxelConeTracing/version.sh
		file = resources/shaders/engine/deferredRendering/gBuffer/gBufferRead.sh
	}

	semantics
	{
		elapsedTime = ELAPSED_TIME
	}
}