shader{

	code= "	
	#define MAX_SAMPLES 64

	#define NOISE_TEXTURE_SIZE	32.0
	#define DEPTH_BIAS			0.00025
	#define MIN_KERNEL_RADIUS	0.03
	#define MAX_KERNEL_RADIUS	0.2

	uniform vec2		occluderSamplingVectors[MAX_SAMPLES];
	uniform vec2		samplingVectors[MAX_SAMPLES];
	
	uniform int			numOccluderSearchSamples;
	uniform int			numSamples;

	uniform sampler2D	noiseTexture;
	uniform float		viewportWidth;
	uniform float		viewportHeight;


	float sampleSoftShadowMap(sampler2D shadowMap, vec4 shadowTexCoord, float kernelRadius, vec2 screenSpacePosition, float depthBias)
	{
		vec2 noiseTextureCoordinates = screenSpacePosition * vec2(viewportWidth, viewportHeight) / NOISE_TEXTURE_SIZE;
		float kernelRotationAngle= texture2D(noiseTexture, noiseTextureCoordinates).x * 2.0 * 3.14;
		mat2 kernelRotationMatrix= mat2(cos(kernelRotationAngle), -sin(kernelRotationAngle), sin(kernelRotationAngle), cos(kernelRotationAngle));
		float result= 0.0;

		for(int i= 0; i< numSamples; ++i)
		{
			float occluderDepth= texture2D(shadowMap, shadowTexCoord.xy+ samplingVectors[i]*kernelRotationMatrix*kernelRadius).x * 0.5 + 0.5;
			
			if(shadowTexCoord.z< occluderDepth+ depthBias)
				result+= 1.0;
		}
		
		return result/numSamples;
	}


	float samplePercentageCloserSoftShadows(sampler2D shadowMap, vec4 shadowTexCoord, vec2 screenSpacePosition, float uvSpaceLightSize, float depthBias = 0.00025)
	{
		//Calculate the textures coordinates for the noise texture so that it's pixels match the pixels of the screen without any resizing.
		vec2 noiseTextureCoordinates = screenSpacePosition * vec2(viewportWidth, viewportHeight) / NOISE_TEXTURE_SIZE;

		//Sample the noise texture and generate a rotation matrix.
		float kernelRotationAngle= texture2D(noiseTexture, noiseTextureCoordinates).x* 2.0 * 3.14;
		mat2 kernelRotationMatrix= mat2(cos(kernelRotationAngle), -sin(kernelRotationAngle), sin(kernelRotationAngle), cos(kernelRotationAngle));

		vec2 texcoord= shadowTexCoord.xy;
		float receiverDepth= shadowTexCoord.z;
		
		//Find average occluder depth.
		//Note: the occluder search radius is fixed because we only use PCSS for sun light. For non paralel light types 
		//the occluder radius should be calculated according to the original documentation of this technnique.
		float searchRadius= uvSpaceLightSize;
		int numOccluders= 0;
		float averageOccluderDepth= 0.0;
		
		for(int i= 0; i< numOccluderSearchSamples; ++i)
		{
			vec2 offset = occluderSamplingVectors[i] * kernelRotationMatrix * searchRadius;
			float occluderDepth= texture2D(shadowMap, texcoord + offset).x * 0.5 + 0.5;
			
			if(receiverDepth > occluderDepth - depthBias)
			{
				averageOccluderDepth += occluderDepth;
				numOccluders++;
			}
		}
		
		averageOccluderDepth = averageOccluderDepth/numOccluders;
		
		//Calculate the kernel radius but don't let it become too small nor too large.
		float penumbraSize = (receiverDepth - averageOccluderDepth) / averageOccluderDepth;
		float kernelRadius = clamp(penumbraSize / receiverDepth, MIN_KERNEL_RADIUS, MAX_KERNEL_RADIUS) * uvSpaceLightSize;

		//Sample the shadow map.
		float shadow = sampleSoftShadowMap(shadowMap, shadowTexCoord, kernelRadius, screenSpacePosition, depthBias);

		return shadow;
	}"
	
	semantics{
		occluderSamplingVectors = SOFT_SHADOW_MAPPING_OCCLUDER_SEARCH_SAMPLING_VECTORS
		samplingVectors= SOFT_SHADOW_MAPPING_SAMPLING_VECTORS
		numOccluderSearchSamples= SOFT_SHADOW_MAPPING_NUM_OCCLUDER_SEARCH_SAMPLES
		numSamples= SOFT_SHADOW_MAPPING_NUM_SAMPLES
		noiseTexture= SHADOW_SAMPLING_NOISE_TEXTURE
		viewportWidth = VIEWPORT_WIDTH
		viewportHeight = VIEWPORT_HEIGHT
	}
}