shader{

	code= "
	uniform float											viewportWidth;
	uniform float											viewportHeight;
	
	layout (rgba16f) writeonly uniform image2D 				frameBuffer;
	layout (local_size_x = 32, local_size_y = 32)			in;


	void main()
	{
		//Get the pixel coordinates.
		uvec2 pixelCoord = gl_GlobalInvocationID.xy;

		//Early out if we're processing outside the viewport limits.
		if(pixelCoord.x >= viewportWidth || pixelCoord.y >= viewportHeight)
		{
			return;
		}

		//Calculate the coordinates and other useful data.
		const vec2 halfTexelOffset = vec2(0.5, 0.5);
		const vec2 viewportDimensions = vec2(viewportWidth, viewportHeight);
		vec2 texCoord = (vec2(pixelCoord) + halfTexelOffset) / viewportDimensions;

		//Get the specular multiplir for this pixel.
		float specularMultiplier = getSpecularMultiplier(texCoord);

		//Only process the reflection cone if the specular intensity is high enough to make the reflection visible.
		if(specularMultiplier > 0.001)
		{
			//Get the rest of the pixel's data.
			vec3 position = reconstructPositionFromDepth(texCoord);
			vec3 normal = getNormal(texCoord);
			vec3 geometryNormal = getGeometryNormal(texCoord);
			float specularRoughness = getSpecularRoughness(texCoord);

			//Find how wide the cone should be based on the surface's roughness. We want
			//our reflection cone to match the specular lobe, so we look at the Cook-Torrance
			//formula that we use to calculate the specular and see that the lobe is described
			//by a simple Gaussian function. The shape of this function is controlled by the
			//roughness factor which is acts as the standard deviation of the function.
			//A property of the Gaussian distribution is that 68% of the function is within the
			//[-deviation, +deviation] so our roughness factor gives us a pretty good estimate
			//of the hotspot angular width of the specular and that's what we use as the cone angle.
			float coneAngle = 2.0 * specularRoughness;

			//Calculate the reflection cone.
			vec3 viewDirection = normalize(position - cameraPosition);
			vec3 reflectionDirection = normalize(reflect(viewDirection, normal));
			vec4 coneResult = traceReflectionCone(position, geometryNormal, reflectionDirection, coneAngle);
			vec3 reflection = coneResult.rgb;

			//Add the sky lighting contribution for the parts that aren't occluded by the scene.
			float skyLightingOcclusion = coneResult.a;
			vec3 skyLighting = sampleSkyMap(reflectionDirection, coneAngle) * skyLightingOcclusion;

			vec3 result = (reflection + skyLighting) * specularMultiplier;

			//Store the result.
			imageStore(frameBuffer, ivec2(pixelCoord), vec4(result, 1));
		}
	}
	"

	includes
	{
		file = resources/shaders/engine/voxelConeTracing/version.sh
		file = resources/shaders/engine/deferredRendering/gBuffer/gBufferRead.sh
		file = resources/shaders/engine/voxelConeTracing/voxelConeTracing.sh
		file = resources/shaders/engine/atmosphere/skyLighting/skyMapRead.sh
	}

	semantics
	{
		viewportWidth = VIEWPORT_WIDTH
		viewportHeight = VIEWPORT_HEIGHT
	}
}