shader{

	code= "

	#define				VOLUME_1	0	
	#define				VOLUME_2	1
	#define				NO_VOLUME	2

	//General constants.
	const float			DIFFUSE_GI_INTENSITY = 0.5;
	const float			SKYLIGHTING_INTENSITY = 0.05;

	//Diffuse GI constants.
	const float			GI_CONE_ANGLE = 0.3;
	const float			GI_CONE_LENGTH = 40.0;
	const int			GI_NUM_SAMPLES = 32;
	const float			GI_MIN_SAMPLING_RADIUS_MULTIPLIER = 3.0;

	//Reflection constants.
	const float			REFLECTION_CONE_LENGTH = 100.0;
	const int			REFLECTION_NUM_SAMPLES = 200;
	const float			MIN_SEARCH_RADIUS = 0.5;
	const float			MAX_SEARCH_RADIUS = 2.0;
	const float			SEARCH_RADIUS_INCREMENT = 0.5;


	//Volume 1 textures.
	uniform sampler3D	voxelTextureNX1;
	uniform sampler3D	voxelTexturePX1;
	uniform sampler3D	voxelTextureNY1;
	uniform sampler3D	voxelTexturePY1;
	uniform sampler3D	voxelTextureNZ1;
	uniform sampler3D	voxelTexturePZ1;

	//Volume 2 textures.
	uniform sampler3D	voxelTextureNX2;
	uniform sampler3D	voxelTexturePX2;
	uniform sampler3D	voxelTextureNY2;
	uniform sampler3D	voxelTexturePY2;
	uniform sampler3D	voxelTextureNZ2;
	uniform sampler3D	voxelTexturePZ2;

	//Volumes dimensions.
	uniform vec3		voxelVolumePosition1;
	uniform vec3		voxelVolumePosition2;
	uniform	float		voxelVolumeSize1;
	uniform	float		voxelVolumeSize2;
	uniform	int			voxelVolumeResolution;


	int getVolumeWherePointIsLocated(vec3 point)
	{
		float halfVolumeSize1 = voxelVolumeSize1 / 2.0;
		float halfVolumeSize2 = voxelVolumeSize2 / 2.0;

		if(point.x >= -halfVolumeSize1 && point.x <= halfVolumeSize1 && point.y >= -halfVolumeSize1 && point.y <= halfVolumeSize1 && point.z >= -halfVolumeSize1 && point.z <= halfVolumeSize1)
		{
			return VOLUME_1;
		}
		else if(point.x >= -halfVolumeSize2 && point.x <= halfVolumeSize2 && point.y >= -halfVolumeSize2 && point.y <= halfVolumeSize2 && point.z >= -halfVolumeSize2 && point.z <= halfVolumeSize2)
		{
			return VOLUME_2;
		}
		else
		{
			return NO_VOLUME;
		}
	}


	vec4 sampleVolume(vec3 position, vec3 direction, float radius, vec3 voxelVolumePosition, float voxelVolumeSize, sampler3D voxelTextureNX, sampler3D voxelTexturePX, sampler3D voxelTextureNY, sampler3D voxelTexturePY, sampler3D voxelTextureNZ, sampler3D voxelTexturePZ)
	{
		vec3 texcoord = (position - voxelVolumePosition).xzy / voxelVolumeSize + 0.5;			
		float lod = log2(2.0 * radius * voxelVolumeResolution / voxelVolumeSize);			//Calculate the lod level.

		direction = -direction;							//Invert the direction to transform it from incidence to outgoing direction.
		float dotNegativeX = max(-direction.x, 0.0);
		float dotPositiveX = max(direction.x, 0.0);
		float dotNegativeY = max(-direction.y, 0.0);
		float dotPositiveY = max(direction.y, 0.0);
		float dotNegativeZ = max(-direction.z, 0.0);
		float dotPositiveZ = max(direction.z, 0.0);

		vec4 result = vec4(0, 0, 0, 0);

		if(dotNegativeX > 0.0)
		{
			result += textureLod(voxelTextureNX, texcoord, lod) * dotNegativeX;
		}

		if(dotPositiveX > 0.0)
		{
			result += textureLod(voxelTexturePX, texcoord, lod) * dotPositiveX;
		}

		if(dotNegativeY > 0.0)
		{
			result += textureLod(voxelTextureNY, texcoord, lod) * dotNegativeY;
		}

		if(dotPositiveY > 0.0)
		{
			result += textureLod(voxelTexturePY, texcoord, lod) * dotPositiveY;
		}

		if(dotNegativeZ > 0.0)
		{
			result += textureLod(voxelTextureNZ, texcoord, lod) * dotNegativeZ;
		}

		if(dotPositiveZ > 0.0)
		{
			result += textureLod(voxelTexturePZ, texcoord, lod) * dotPositiveZ;
		}
		
		return result;
	}

	
	vec4 sampleDirectionalVoxel(vec3 position, vec3 direction, float radius)
	{	
		//Check in which volume is the point inside of and sample from it accordingly.
		switch(getVolumeWherePointIsLocated(position))
		{
			case VOLUME_1:	return sampleVolume(position, direction, radius, voxelVolumePosition1, voxelVolumeSize1, voxelTextureNX1, voxelTexturePX1, voxelTextureNY1, voxelTexturePY1, voxelTextureNZ1, voxelTexturePZ1);
			case VOLUME_2:	return sampleVolume(position, direction, radius, voxelVolumePosition2, voxelVolumeSize2, voxelTextureNX2, voxelTexturePX2, voxelTextureNY2, voxelTexturePY2, voxelTextureNZ2, voxelTexturePZ2);
			case NO_VOLUME:	return vec4(0.0);
		}
	}


	vec4 traceCone(vec3 position, vec3 surfaceNormal, vec3 coneDirection, float coneAngle, float coneLength, int numSamples, float minSamplingRadiusMultiplier, bool isReflectionCone)
	{
		//Perform some pre-calculations.
		float coneHalfAngle = coneAngle / 2.0;
		float tangent = tan(coneHalfAngle);
		float distanceBetweenSamples = coneLength / numSamples;
		
		//Calculate the voxel size of each volume.
		float volumeVoxelSize1 = (voxelVolumeSize1 / voxelVolumeResolution);
		float volumeVoxelSize2 = (voxelVolumeSize2 / voxelVolumeResolution);

		//Find the volume from which the cone starts and store its voxel size.
		int startingVolume = getVolumeWherePointIsLocated(position);
		float startingVolumeVoxelSize = (startingVolume == VOLUME_1) ? volumeVoxelSize1 : volumeVoxelSize2;
		
		//Offset the initial position in the direction of the surface normal (attention, not the bumped normal)
		//and by the distance of at least one voxel to avoid self-intersecting the surface the cone originates from.
		position = position + (surfaceNormal + coneDirection) * startingVolumeVoxelSize;

		//Trace the cone.
		float distance = 0;
		float accumulatedOpacity = 0;
		vec4 previousSample = vec4(0, 0, 0, 0);
		float distanceToPreviousSample = 0;
		vec3 rayColor = vec3(0, 0, 0);

		float searchRadius = MIN_SEARCH_RADIUS;

		for(int i = 0; i < numSamples; i++)
		{
			vec3 samplePosition = position + coneDirection * distance;
			float radius = distance * tangent;

			//For the diffuse GI we need to use a minimum sampling radius based on the voxel size to
			//ensure the results are soft when the cone intersects a surface near its origin position.
			radius = max(radius, startingVolumeVoxelSize / 2.0 * minSamplingRadiusMultiplier);

			//For the reflections it's important to skip empty spaces to improve the performance.
			//To detect empty spaces we sample the volume using a fairly large sampling radius and
			//look at the resulting opacity to check if there's anything in the surroundings of the
			//sample, if not then we can skip the entire free space.
			//We can improve this by using a variable search radius. Everytime we catch an empty space
			//we increase the search radius a bit in an attempt to catch an even larger empty space in
			//the next iteration, but if we get a non-empty space then we decrease the sampling radius
			//progressively to narrow the search and make the sample reach the deetcted obstacle quickly.
			//Although other smarter tricks could be devised based on sampling with a larger radius they
			//do not always yield a performance improvement. There's a sentitive balance the amount of
			//memory accesses, the coeherency of those accesses and the amount of branching performed on
			//the shader that prevents us from getting performance improvements from more complex algorithms
			//(like recursively traversing the mipmap chain).
			if(isReflectionCone)
			{
				//Take a peek around the sample to see if there's any opacity.
				float hintOpacity = sampleDirectionalVoxel(samplePosition, coneDirection, searchRadius).a;

				//If there's nothing around the sample.
				if(hintOpacity < 0.01)
				{
					//Increment the sample distance by the search radius.
					distance += searchRadius;

					//Increment the search radius but make sure it doesn't become too large which would be counterproductive.
					searchRadius = min(searchRadius + SEARCH_RADIUS_INCREMENT, MAX_SEARCH_RADIUS);

					//Proceed to the next iteration.
					continue;
				}
				else
				{
					//Decrease the search radius but prevent it from becoming too small.
					searchRadius = max(searchRadius - SEARCH_RADIUS_INCREMENT, MIN_SEARCH_RADIUS);
				}
			}

			//Sample the voxel volume.
			vec4 currentSample = sampleDirectionalVoxel(samplePosition, coneDirection, radius);

			//The color/opacity of the sample is merely a discrete value, we need to integrate them across
			//the length of the sample, as described in the paper 'Optical Models for Direct Volume Rendering'
			//by Nelson Max, to obtain the contribution of this sample to the total color and opacity.
			float totalSampleOpacity = 1.0 - exp(-currentSample.a * distanceToPreviousSample);
			vec3 totalSampleColor = currentSample.rgb / (currentSample.a + 0.00001) * totalSampleOpacity;

			//HACK: for the reflections we'll use the original sample values instead of the
			//integrated ones because they generate a more accurate reflection of the scene.
			if(isReflectionCone)
			{
				totalSampleColor = currentSample.rgb;
				totalSampleOpacity = currentSample.a;
			}

			//Accumulate the color and opacity.
			rayColor += totalSampleColor * (1.0 - accumulatedOpacity);
			accumulatedOpacity += (1.0 - accumulatedOpacity) * totalSampleOpacity;

			//Increment the sample distance to the next step, this is done differently for the diffuse and specular lighting.
			if(isReflectionCone)
			{
				//For the reflections, specially the sharp ones, we need to keep the samples tight to ensure
				//that we don't miss any scene features but also avoid sampling at too small intervals so we
				//won't need an insage amount of samples to cover the entire scene.
				//We do this by assuming that samples are spheres. So the radius of the next sample must
				//be choosen so that the next sample sphere touches the current sample sphare without
				//intersecting it and at the same time fit tighly inside the cone. This can be achieved
				//through fairly simple trigonometry derivation which gives us the optimal radius for the
				//next samples that obeys the previous requirements.
				float nextSampleRadius = (radius + radius*radius/ distance) / (1 - radius/distance);
				
				//Avoid sampling the same voxel by ensuring that the step is at least half voxel long.
				float distanceStep = max(nextSampleRadius, volumeVoxelSize1 * 0.5);

				//Increment the distance.
				distance += distanceStep;
				distanceToPreviousSample = distanceStep;
			}
			else
			{
				//For the diffuse lighting we get better results by doing regular sampling of the cone.
				distance += distanceBetweenSamples;
				distanceToPreviousSample = distanceBetweenSamples;
			}

			//Store the sample to be used in the next iteration.
			previousSample = currentSample;

			//Break out if we reached full opacity or exceeded the cone length.
			if(accumulatedOpacity >= 1.0 || distance >= coneLength)
			{
				accumulatedOpacity = min(accumulatedOpacity, 1.0); //Clamp the accumulated opacity to 1.0
				break;
			}
		}

		//Return the color and the accumulated opacity.
		return vec4(rayColor, 1.0 - accumulatedOpacity);
	}

	vec4 traceDiffuseCone(vec3 position, vec3 surfaceNormal, vec3 coneDirection)
	{
		return traceCone(position, surfaceNormal, coneDirection, GI_CONE_ANGLE, GI_CONE_LENGTH, GI_NUM_SAMPLES, GI_MIN_SAMPLING_RADIUS_MULTIPLIER, false);
	}

	vec4 traceReflectionCone(vec3 position, vec3 surfaceNormal, vec3 coneDirection, float coneAngle)
	{
		return traceCone(position, surfaceNormal, coneDirection, coneAngle, REFLECTION_CONE_LENGTH, REFLECTION_NUM_SAMPLES, 0.0, true);
	}
	"


	semantics
	{
		voxelTextureNX1 = VOXEL_VOLUME_1_TEXTURE_NX
		voxelTexturePX1 = VOXEL_VOLUME_1_TEXTURE_PX
		voxelTextureNY1 = VOXEL_VOLUME_1_TEXTURE_NY
		voxelTexturePY1 = VOXEL_VOLUME_1_TEXTURE_PY
		voxelTextureNZ1 = VOXEL_VOLUME_1_TEXTURE_NZ
		voxelTexturePZ1 = VOXEL_VOLUME_1_TEXTURE_PZ
		voxelTextureNX2 = VOXEL_VOLUME_2_TEXTURE_NX
		voxelTexturePX2 = VOXEL_VOLUME_2_TEXTURE_PX
		voxelTextureNY2 = VOXEL_VOLUME_2_TEXTURE_NY
		voxelTexturePY2 = VOXEL_VOLUME_2_TEXTURE_PY
		voxelTextureNZ2 = VOXEL_VOLUME_2_TEXTURE_NZ
		voxelTexturePZ2 = VOXEL_VOLUME_2_TEXTURE_PZ
		voxelVolumePosition1 = VOXEL_VOLUME_1_POSITION
		voxelVolumePosition2 = VOXEL_VOLUME_2_POSITION
		voxelVolumeSize1 = VOXEL_VOLUME_1_SIZE
		voxelVolumeSize2 = VOXEL_VOLUME_2_SIZE
		voxelVolumeResolution = VOXEL_VOLUME_RESOLUTION
	}
}