shader
{
	code= "
		#version 420
		#extension GL_EXT_geometry_shader4 : enable

		uniform mat4	objectMatrixGS;
		uniform mat4	transformationMatrixX;
		uniform mat4	transformationMatrixY;
		uniform mat4	transformationMatrixZ;
		uniform int		voxelVolumeResolution;

		in vec3			position2[3];
		in vec3			normal2[3];
		in vec3			tangent2[3];
		in vec3			binormal2[3];
		in vec2			texCoord2[3];

		out vec3		position3;
		out vec3		normal3;
		out vec3		tangent3;
		out vec3		binormal3;
		out vec2		texCoord3;
		out vec3		positionInVolume;
		out vec2		projectedPosition;

		out vec2		clippingRectangleMin;
		out vec2		clippingRectangleMax;


		struct Edge
		{
			vec2 start;
			vec2 direction;
			vec2 normal;
		};


		Edge createEdge(vec2 vertex1, vec2 vertex2)
		{
			Edge edge;
			edge.start = vertex1;
			edge.direction = normalize(vertex2 - vertex1);
			edge.normal = vec2(edge.direction.y, -edge.direction.x);

			return edge;
		}


		vec2 intersectEdges(Edge edge1, Edge edge2)
		{
			float u = ((edge2.start.y - edge1.start.y) * edge2.direction.x - (edge2.start.x - edge1.start.x) * edge2.direction.y) / (edge2.direction.x * edge1.direction.y - edge2.direction.y * edge1.direction.x);
			return edge1.start + edge1.direction * u;
		}


		vec2 calculateBarycentricCoordinates(vec2 point, vec2 vertex1, vec2 vertex2, vec2 vertex3)
		{
			//This code was extracted from Adrian Boeing's article about barycentric coordinates which can
			//be found on his blog: http://adrianboeing.blogspot.pt/2010/01/barycentric-coordinates.html.

			//Compute vectors.
			vec2 v0 = vertex3 - vertex1;
			vec2 v1 = vertex2 - vertex1;
			vec2 v2 = point - vertex1;

			//Compute dot products.
			float dot00 = dot(v0, v0);
			float dot01 = dot(v0, v1);
			float dot02 = dot(v0, v2);
			float dot11 = dot(v1, v1);
			float dot12 = dot(v1, v2);

			//Compute barycentric coordinates.
			float invDenom = 1.0 / (dot00 * dot11 - dot01 * dot01);
			float u = (dot11 * dot02 - dot01 * dot12) * invDenom;
			float v = (dot00 * dot12 - dot01 * dot02) * invDenom;

			return vec2(u, v);
		}


		void main()
		{
			vec4 projectedVertexPositions[3];
			vec3 verticesPositionsInVolume[3];
			vec2 clippingMin = vec2(1, 1); 
			vec2 clippingMax = vec2(-1, -1);

			//Calculate the triangle normal with its components converted to positive values so we can use it to find the dominant axis.
			vec3 triangleEdgeVector1 = gl_PositionIn[1].xyz - gl_PositionIn[0].xyz;
			vec3 triangleEdgevec2 = gl_PositionIn[2].xyz - gl_PositionIn[0].xyz;
			vec3 absoluteTriangleNormal = abs(normalize(cross(triangleEdgeVector1, triangleEdgevec2)));

			//Find the dominant axis and select the corresponding projection matrix ===========================
			int dominantAxis;
			mat4 projectionMatrix;

			if(absoluteTriangleNormal.x >= absoluteTriangleNormal.y && absoluteTriangleNormal.x >= absoluteTriangleNormal.z)
			{
				dominantAxis = 0;
				projectionMatrix = transformationMatrixX;
			}
			else if(absoluteTriangleNormal.y >= absoluteTriangleNormal.x && absoluteTriangleNormal.y >= absoluteTriangleNormal.z)
			{
				dominantAxis = 1;
				projectionMatrix = transformationMatrixY;
			}
			else
			{
				dominantAxis = 2;
				projectionMatrix = transformationMatrixZ;
			}
			//=================================================================================================



			//Project the vertices and update the clipping rectangle ==========================================
			for(int i = 0; i < 3; i++)
			{
				//Project the vertex.
				projectedVertexPositions[i] = projectionMatrix * gl_PositionIn[i];;

				//Update the clipping rectangle.
				clippingMin.x = min(clippingMin.x, projectedVertexPositions[i].x);
				clippingMax.x = max(clippingMax.x, projectedVertexPositions[i].x);
				clippingMin.y = min(clippingMin.y, projectedVertexPositions[i].y);
				clippingMax.y = max(clippingMax.y, projectedVertexPositions[i].y);
			}

			//Convert the clipping rectangle to normalized coordinates, expand the clipping
			//rectangle by one pixel in each direction and send it to the pixel shader.
			const float pixelSize = 1.0 / voxelVolumeResolution;
			clippingMin = clippingMin * 0.5 + 0.5;
			clippingMax = clippingMax * 0.5 + 0.5;
			clippingRectangleMin = clippingMin - vec2(pixelSize, pixelSize);
			clippingRectangleMax = clippingMax + vec2(pixelSize, pixelSize);
			//=================================================================================================



			//Enlarge the triangle ============================================================================
			const float offset = 2.0 / voxelVolumeResolution / 2.0;
			
			Edge edge1 = createEdge(projectedVertexPositions[0].xy, projectedVertexPositions[1].xy);
			Edge edge2 = createEdge(projectedVertexPositions[1].xy, projectedVertexPositions[2].xy);
			Edge edge3 = createEdge(projectedVertexPositions[2].xy, projectedVertexPositions[0].xy);

			//Invert the edge normals if the face is back facing
			vec3 projectedEdge1 = vec3(edge1.direction.x, 0.0, edge1.direction.y);
			vec3 projectedEdge2 = vec3(edge2.direction.x, 0.0, edge2.direction.y);
			bool isBackfacing = cross(projectedEdge1, projectedEdge2).y > 0.0f;

			if(isBackfacing)
			{
				edge1.normal = -edge1.normal;
				edge2.normal = -edge2.normal;
				edge3.normal = -edge3.normal;
			}

			//Enlarge the triangle by moving its edges away.
			edge1.start += edge1.normal * offset;
			edge2.start += edge2.normal * offset;
			edge3.start += edge3.normal * offset;

			//Calculate the new vertex positions by calculating the intersection points between the expanded edges.
			vec4 expandedVertexPositions[3] = projectedVertexPositions;
			expandedVertexPositions[0].xy = intersectEdges(edge1, edge3);
			expandedVertexPositions[1].xy = intersectEdges(edge1, edge2);
			expandedVertexPositions[2].xy = intersectEdges(edge2, edge3);
			//=================================================================================================


			//Adjust the texture coordinates to compensate for the triangle scaling ===========================

			//The idea behind the following implementation is this: 
			// - While the projected vertices define a triangle in screen space the texture coordinates define the same triangle in texture space
			// - When the vertices are expanded in screen space they must be proportionaly expanded in texture space
			// - Therefore we must find how much have the vertices expanded in screen space and apply the same expansion to the texture coordinates
			// - The barycentric coordinates of the expanded vertices using the original triangle as reference gives us how much have they expanded
			// - Using the same barycentric coordinates in texture space gives us the expanded texture coordinates

			
			//Calculate the barycentric coordinates of the expanded vertex positions on the original triangle.
			vec2 vertex1BarycentricCoords = calculateBarycentricCoordinates(expandedVertexPositions[0].xy, projectedVertexPositions[0].xy, projectedVertexPositions[1].xy, projectedVertexPositions[2].xy);
			vec2 vertex2BarycentricCoords = calculateBarycentricCoordinates(expandedVertexPositions[1].xy, projectedVertexPositions[0].xy, projectedVertexPositions[1].xy, projectedVertexPositions[2].xy);
			vec2 vertex3BarycentricCoords = calculateBarycentricCoordinates(expandedVertexPositions[2].xy, projectedVertexPositions[0].xy, projectedVertexPositions[1].xy, projectedVertexPositions[2].xy);

			//Calculate the texture coordinates vectors for two of the triangle edges.
			vec2 texCoordVector1 = (texCoord2[2] - texCoord2[0]);
			vec2 texCoordVector2 = (texCoord2[1] - texCoord2[0]);

			//Calculate the new texture coordinates.
			vec2 textureCoordinates[3];
			textureCoordinates[0] = texCoord2[0] + texCoordVector1 * vertex1BarycentricCoords.x + texCoordVector2 * vertex1BarycentricCoords.y;
			textureCoordinates[1] = texCoord2[0] + texCoordVector1 * vertex2BarycentricCoords.x + texCoordVector2 * vertex2BarycentricCoords.y;
			textureCoordinates[2] = texCoord2[0] + texCoordVector1 * vertex3BarycentricCoords.x + texCoordVector2 * vertex3BarycentricCoords.y;
			//=================================================================================================


			
			//Emit the new triangle vertices ==================================================================
			for(int i = 0; i < 3; i++)
			{
				//Obtain the vertex positions in volume space.
				switch(dominantAxis)
				{
					case 0:	verticesPositionsInVolume[i] = expandedVertexPositions[i].zxy * 0.5 + 0.5;
							break;

					case 1:	verticesPositionsInVolume[i] = expandedVertexPositions[i].xyz * 0.5 + 0.5;
							break;

					case 2:	verticesPositionsInVolume[i] = expandedVertexPositions[i].xzy * 0.5 + 0.5;
							verticesPositionsInVolume[i].x = 1.0 - verticesPositionsInVolume[i].x;
							break;
				}

				//Output the vertex data.
				gl_Position = expandedVertexPositions[i];
				position3 = gl_PositionIn[i].xyz;
				positionInVolume = verticesPositionsInVolume[i];
				projectedPosition = expandedVertexPositions[i].xy * 0.5 + 0.5;
				normal3 = normal2[i];
				tangent3 = tangent2[i];
				binormal3 = binormal2[i];
				texCoord3 = textureCoordinates[i];
				EmitVertex();
			}
			//=================================================================================================
		}
	"

	semantics
	{
		objectMatrixGS = OBJECT_TRANSFORM_MATRIX
		voxelVolumeResolution = VOXEL_VOLUME_RESOLUTION
	}
}